/*
Robot36

Copyright 2024 Ahmet Inan <xdsopl@gmail.com>
*/

package xdsopl.robot36;

import android.Manifest;
import android.content.ContentResolver;
import android.content.ContentValues;
import android.content.Context;
import android.content.Intent;
import android.content.SharedPreferences;
import android.content.pm.PackageManager;
import android.content.res.Configuration;
import android.graphics.Bitmap;
import android.media.AudioFormat;
import android.media.AudioRecord;
import android.media.MediaRecorder;
import android.net.Uri;
import android.os.Build;
import android.os.Bundle;
import android.os.Environment;
import android.os.ParcelFileDescriptor;
import android.provider.MediaStore;
import android.text.Html;
import android.text.method.LinkMovementMethod;
import android.util.DisplayMetrics;
import android.view.Gravity;
import android.view.LayoutInflater;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.widget.ImageView;
import android.widget.TextView;
import android.widget.Toast;

import androidx.activity.EdgeToEdge;
import androidx.annotation.NonNull;
import androidx.appcompat.app.AlertDialog;
import androidx.appcompat.app.AppCompatActivity;
import androidx.appcompat.app.AppCompatDelegate;
import androidx.appcompat.widget.ShareActionProvider;
import androidx.core.app.ActivityCompat;
import androidx.core.content.ContextCompat;
import androidx.core.graphics.Insets;
import androidx.core.view.MenuItemCompat;
import androidx.core.view.ViewCompat;
import androidx.core.view.WindowInsetsCompat;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.Locale;

public class MainActivity extends AppCompatActivity {

	private Bitmap scopeBitmap;
	private PixelBuffer scopeBuffer;
	private ImageView scopeView;
	private Bitmap freqPlotBitmap;
	private PixelBuffer freqPlotBuffer;
	private ImageView freqPlotView;
	private Bitmap peakMeterBitmap;
	private PixelBuffer peakMeterBuffer;
	private ImageView peakMeterView;
	private PixelBuffer imageBuffer;
	private float[] recordBuffer;
	private AudioRecord audioRecord;
	private Decoder decoder;
	private Menu menu;
	private String forceMode;
	private int recordRate;
	private int recordChannel;
	private int audioSource;
	private int fgColor;
	private int thinColor;
	private int tintColor;

	private void setStatus(int id) {
		setTitle(id);
	}

	private void setStatus(String str) {
		setTitle(str);
	}

	private void forceMode(int id) {
		menu.findItem(R.id.action_auto_mode).setIcon(R.drawable.baseline_lock_24);
		forceMode = getString(id);
		if (decoder != null)
			decoder.forceMode(forceMode);
	}

	private void autoMode() {
		int icon;
		if (decoder == null || forceMode != null && !forceMode.equals(getString(R.string.auto_mode))) {
			icon = R.drawable.baseline_auto_mode_24;
			forceMode = getString(R.string.auto_mode);
		} else {
			icon = R.drawable.baseline_lock_24;
			forceMode = decoder.currentMode.getName();
		}
		menu.findItem(R.id.action_auto_mode).setIcon(icon);
		if (decoder != null)
			decoder.forceMode(forceMode);
	}

	private final AudioRecord.OnRecordPositionUpdateListener recordListener = new AudioRecord.OnRecordPositionUpdateListener() {
		@Override
		public void onMarkerReached(AudioRecord ignore) {
		}

		@Override
		public void onPeriodicNotification(AudioRecord audioRecord) {
			audioRecord.read(recordBuffer, 0, recordBuffer.length, AudioRecord.READ_BLOCKING);
			processPeakMeter();
			boolean newLines = decoder.process(recordBuffer, recordChannel);
			processFreqPlot();
			if (newLines) {
				processScope();
				processImage();
				setStatus(decoder.currentMode.getName());
			}
		}
	};

	private void processPeakMeter() {
		float max = 0;
		for (float v : recordBuffer)
			max = Math.max(max, Math.abs(v));
		int pixels = peakMeterBuffer.height;
		int peak = pixels;
		if (max > 0)
			peak = (int) Math.round(Math.min(Math.max(-Math.PI * Math.log(max), 0), pixels));
		Arrays.fill(peakMeterBuffer.pixels, 0, peak, thinColor);
		Arrays.fill(peakMeterBuffer.pixels, peak, pixels, tintColor);
		peakMeterBitmap.setPixels(peakMeterBuffer.pixels, 0, peakMeterBuffer.width, 0, 0, peakMeterBuffer.width, peakMeterBuffer.height);
		peakMeterView.invalidate();
	}

	private void processFreqPlot() {
		int width = freqPlotBitmap.getWidth();
		int height = freqPlotBitmap.getHeight();
		int stride = freqPlotBuffer.width;
		int line = stride * freqPlotBuffer.line;
		int channels = recordChannel > 0 ? 2 : 1;
		int samples = recordBuffer.length / channels;
		int spread = 2;
		Arrays.fill(freqPlotBuffer.pixels, line, line + stride, 0);
		for (int i = 0; i < samples; ++i) {
			int x = Math.round((recordBuffer[i] + 2.5f) * 0.25f * stride);
			if (x >= spread && x < stride - spread)
				for (int j = -spread; j <= spread; ++j)
					freqPlotBuffer.pixels[line + x + j] += 1 + spread * spread - j * j;
		}
		int factor = 960 / samples;
		for (int i = 0; i < stride; ++i)
			freqPlotBuffer.pixels[line + i] = 0x00FFFFFF & fgColor | Math.min(factor * freqPlotBuffer.pixels[line + i], 255) << 24;
		System.arraycopy(freqPlotBuffer.pixels, line, freqPlotBuffer.pixels, line + stride * (freqPlotBuffer.height / 2), stride);
		freqPlotBuffer.line = (freqPlotBuffer.line + 1) % (freqPlotBuffer.height / 2);
		int offset = stride * (freqPlotBuffer.line + freqPlotBuffer.height / 2 - height);
		freqPlotBitmap.setPixels(freqPlotBuffer.pixels, offset, stride, 0, 0, width, height);
		freqPlotView.invalidate();
	}

	private void processScope() {
		int width = scopeBitmap.getWidth();
		int height = scopeBitmap.getHeight();
		int stride = scopeBuffer.width;
		int offset = stride * (scopeBuffer.line + scopeBuffer.height / 2 - height);
		scopeBitmap.setPixels(scopeBuffer.pixels, offset, stride, 0, 0, width, height);
		scopeView.invalidate();
	}

	private void processImage() {
		if (imageBuffer.line < imageBuffer.height)
			return;
		imageBuffer.line = -1;
		storeBitmap(Bitmap.createBitmap(imageBuffer.pixels, imageBuffer.width, imageBuffer.height, Bitmap.Config.ARGB_8888));
	}

	private void initAudioRecord() {
		boolean rateChanged = true;
		if (audioRecord != null) {
			rateChanged = audioRecord.getSampleRate() != recordRate;
			boolean channelChanged = audioRecord.getChannelCount() != (recordChannel == 0 ? 1 : 2);
			boolean sourceChanged = audioRecord.getAudioSource() != audioSource;
			if (!rateChanged && !channelChanged && !sourceChanged)
				return;
			stopListening();
			audioRecord.release();
			audioRecord = null;
		}
		int channelConfig = AudioFormat.CHANNEL_IN_MONO;
		int channelCount = 1;
		if (recordChannel != 0) {
			channelCount = 2;
			channelConfig = AudioFormat.CHANNEL_IN_STEREO;
		}
		int sampleSize = 4;
		int frameSize = sampleSize * channelCount;
		int audioFormat = AudioFormat.ENCODING_PCM_FLOAT;
		int readsPerSecond = 50;
		int bufferSize = Integer.highestOneBit(recordRate) * frameSize;
		int frameCount = recordRate / readsPerSecond;
		recordBuffer = new float[frameCount * channelCount];
		try {
			audioRecord = new AudioRecord(audioSource, recordRate, channelConfig, audioFormat, bufferSize);
			if (audioRecord.getState() == AudioRecord.STATE_INITIALIZED) {
				audioRecord.setRecordPositionUpdateListener(recordListener);
				audioRecord.setPositionNotificationPeriod(frameCount);
				if (rateChanged) {
					decoder = new Decoder(scopeBuffer, imageBuffer, recordRate);
					decoder.forceMode(forceMode);
				}
				startListening();
			} else {
				setStatus(R.string.audio_init_failed);
			}
		} catch (IllegalArgumentException e) {
			setStatus(R.string.audio_setup_failed);
		} catch (SecurityException e) {
			setStatus(R.string.audio_permission_denied);
		}
	}

	private void startListening() {
		if (audioRecord != null) {
			audioRecord.startRecording();
			if (audioRecord.getRecordingState() == AudioRecord.RECORDSTATE_RECORDING) {
				audioRecord.read(recordBuffer, 0, recordBuffer.length, AudioRecord.READ_BLOCKING);
				setStatus(R.string.listening);
			} else {
				setStatus(R.string.audio_recording_error);
			}
		}
	}

	private void stopListening() {
		if (audioRecord != null)
			audioRecord.stop();
	}

	private void setRecordRate(int newSampleRate) {
		if (recordRate == newSampleRate)
			return;
		recordRate = newSampleRate;
		updateRecordRateMenu();
		initAudioRecord();
	}

	private void setRecordChannel(int newChannelSelect) {
		if (recordChannel == newChannelSelect)
			return;
		recordChannel = newChannelSelect;
		updateRecordChannelMenu();
		initAudioRecord();
	}

	private void setAudioSource(int newAudioSource) {
		if (audioSource == newAudioSource)
			return;
		audioSource = newAudioSource;
		updateAudioSourceMenu();
		initAudioRecord();
	}

	private void updateRecordRateMenu() {
		switch (recordRate) {
			case 8000:
				menu.findItem(R.id.action_set_record_rate_8000).setChecked(true);
				break;
			case 16000:
				menu.findItem(R.id.action_set_record_rate_16000).setChecked(true);
				break;
			case 32000:
				menu.findItem(R.id.action_set_record_rate_32000).setChecked(true);
				break;
			case 44100:
				menu.findItem(R.id.action_set_record_rate_44100).setChecked(true);
				break;
			case 48000:
				menu.findItem(R.id.action_set_record_rate_48000).setChecked(true);
				break;
		}
	}

	private void updateRecordChannelMenu() {
		switch (recordChannel) {
			case 0:
				menu.findItem(R.id.action_set_record_channel_default).setChecked(true);
				break;
			case 1:
				menu.findItem(R.id.action_set_record_channel_first).setChecked(true);
				break;
			case 2:
				menu.findItem(R.id.action_set_record_channel_second).setChecked(true);
				break;
			case 3:
				menu.findItem(R.id.action_set_record_channel_summation).setChecked(true);
				break;
			case 4:
				menu.findItem(R.id.action_set_record_channel_analytic).setChecked(true);
				break;
		}
	}

	private void updateAudioSourceMenu() {
		switch (audioSource) {
			case MediaRecorder.AudioSource.DEFAULT:
				menu.findItem(R.id.action_set_source_default).setChecked(true);
				break;
			case MediaRecorder.AudioSource.MIC:
				menu.findItem(R.id.action_set_source_microphone).setChecked(true);
				break;
			case MediaRecorder.AudioSource.CAMCORDER:
				menu.findItem(R.id.action_set_source_camcorder).setChecked(true);
				break;
			case MediaRecorder.AudioSource.VOICE_RECOGNITION:
				menu.findItem(R.id.action_set_source_voice_recognition).setChecked(true);
				break;
			case MediaRecorder.AudioSource.UNPROCESSED:
				menu.findItem(R.id.action_set_source_unprocessed).setChecked(true);
				break;
		}
	}

	private final int permissionID = 1;

	@Override
	public void onRequestPermissionsResult(int requestCode, @NonNull String[] permissions, @NonNull int[] grantResults) {
		super.onRequestPermissionsResult(requestCode, permissions, grantResults);
		if (requestCode != permissionID)
			return;
		for (int i = 0; i < permissions.length; ++i)
			if (permissions[i].equals(Manifest.permission.RECORD_AUDIO) && grantResults[i] == PackageManager.PERMISSION_GRANTED)
				initAudioRecord();
	}

	@Override
	protected void onSaveInstanceState(@NonNull Bundle state) {
		state.putInt("nightMode", AppCompatDelegate.getDefaultNightMode());
		state.putInt("recordRate", recordRate);
		state.putInt("recordChannel", recordChannel);
		state.putInt("audioSource", audioSource);
		super.onSaveInstanceState(state);
	}

	private void storeSettings() {
		SharedPreferences pref = getPreferences(Context.MODE_PRIVATE);
		SharedPreferences.Editor edit = pref.edit();
		edit.putInt("nightMode", AppCompatDelegate.getDefaultNightMode());
		edit.putInt("recordRate", recordRate);
		edit.putInt("recordChannel", recordChannel);
		edit.putInt("audioSource", audioSource);
		edit.apply();
	}

	@Override
	protected void onCreate(Bundle state) {
		final int defaultSampleRate = 44100;
		final int defaultChannelSelect = 0;
		final int defaultAudioSource = MediaRecorder.AudioSource.MIC;
		if (state == null) {
			SharedPreferences pref = getPreferences(Context.MODE_PRIVATE);
			AppCompatDelegate.setDefaultNightMode(pref.getInt("nightMode", AppCompatDelegate.getDefaultNightMode()));
			recordRate = pref.getInt("recordRate", defaultSampleRate);
			recordChannel = pref.getInt("recordChannel", defaultChannelSelect);
			audioSource = pref.getInt("audioSource", defaultAudioSource);
		} else {
			AppCompatDelegate.setDefaultNightMode(state.getInt("nightMode", AppCompatDelegate.getDefaultNightMode()));
			recordRate = state.getInt("recordRate", defaultSampleRate);
			recordChannel = state.getInt("recordChannel", defaultChannelSelect);
			audioSource = state.getInt("audioSource", defaultAudioSource);
		}
		super.onCreate(state);
		Configuration config = getResources().getConfiguration();
		EdgeToEdge.enable(this);
		setContentView(config.orientation == Configuration.ORIENTATION_LANDSCAPE ? R.layout.activity_main_land : R.layout.activity_main);
		handleInsets();
		fgColor = getColor(R.color.fg);
		thinColor = getColor(R.color.thin);
		tintColor = getColor(R.color.tint);
		scopeBuffer = new PixelBuffer(640, 2 * 1280);
		createScope(config);
		freqPlotBuffer = new PixelBuffer(256, 2 * 256);
		createFreqPlot(config);
		peakMeterBuffer = new PixelBuffer(1, 16);
		createPeakMeter();
		imageBuffer = new PixelBuffer(640, 496);
		List<String> permissions = new ArrayList<>();
		if (ContextCompat.checkSelfPermission(this, Manifest.permission.RECORD_AUDIO) != PackageManager.PERMISSION_GRANTED) {
			permissions.add(Manifest.permission.RECORD_AUDIO);
			setStatus(R.string.audio_permission_denied);
		} else {
			initAudioRecord();
		}
		if (Build.VERSION.SDK_INT <= Build.VERSION_CODES.P && ContextCompat.checkSelfPermission(this, Manifest.permission.WRITE_EXTERNAL_STORAGE) != PackageManager.PERMISSION_GRANTED)
			permissions.add(Manifest.permission.WRITE_EXTERNAL_STORAGE);
		if (!permissions.isEmpty())
			ActivityCompat.requestPermissions(this, permissions.toArray(new String[0]), permissionID);
	}

	private void handleInsets() {
		ViewCompat.setOnApplyWindowInsetsListener(findViewById(R.id.main), (v, insets) -> {
			Insets systemBars = insets.getInsets(WindowInsetsCompat.Type.systemBars());
			v.setPadding(systemBars.left, systemBars.top, systemBars.right, systemBars.bottom);
			return insets;
		});
	}

	@Override
	public boolean onCreateOptionsMenu(Menu menu) {
		getMenuInflater().inflate(R.menu.menu_main, menu);
		this.menu = menu;
		updateRecordRateMenu();
		updateRecordChannelMenu();
		updateAudioSourceMenu();
		return true;
	}

	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
		int id = item.getItemId();
		if (id == R.id.action_store_scope) {
			storeScope();
			return true;
		}
		if (id == R.id.action_auto_mode) {
			autoMode();
			return true;
		}
		if (id == R.id.action_force_raw_mode) {
			forceMode(R.string.raw_mode);
			return true;
		}
		if (id == R.id.action_force_robot36_color) {
			forceMode(R.string.robot36_color);
			return true;
		}
		if (id == R.id.action_force_robot72_color) {
			forceMode(R.string.robot72_color);
			return true;
		}
		if (id == R.id.action_force_pd50) {
			forceMode(R.string.pd50);
			return true;
		}
		if (id == R.id.action_force_pd90) {
			forceMode(R.string.pd90);
			return true;
		}
		if (id == R.id.action_force_pd120) {
			forceMode(R.string.pd120);
			return true;
		}
		if (id == R.id.action_force_pd160) {
			forceMode(R.string.pd160);
			return true;
		}
		if (id == R.id.action_force_pd180) {
			forceMode(R.string.pd180);
			return true;
		}
		if (id == R.id.action_force_pd240) {
			forceMode(R.string.pd240);
			return true;
		}
		if (id == R.id.action_force_martin1) {
			forceMode(R.string.martin1);
			return true;
		}
		if (id == R.id.action_force_martin2) {
			forceMode(R.string.martin2);
			return true;
		}
		if (id == R.id.action_force_scottie1) {
			forceMode(R.string.scottie1);
			return true;
		}
		if (id == R.id.action_force_scottie2) {
			forceMode(R.string.scottie2);
			return true;
		}
		if (id == R.id.action_force_scottie_dx) {
			forceMode(R.string.scottie_dx);
			return true;
		}
		if (id == R.id.action_force_wraase_sc2_180) {
			forceMode(R.string.wraase_sc2_180);
			return true;
		}
		if (id == R.id.action_set_record_rate_8000) {
			setRecordRate(8000);
			return true;
		}
		if (id == R.id.action_set_record_rate_16000) {
			setRecordRate(16000);
			return true;
		}
		if (id == R.id.action_set_record_rate_32000) {
			setRecordRate(32000);
			return true;
		}
		if (id == R.id.action_set_record_rate_44100) {
			setRecordRate(44100);
			return true;
		}
		if (id == R.id.action_set_record_rate_48000) {
			setRecordRate(48000);
			return true;
		}
		if (id == R.id.action_set_record_channel_default) {
			setRecordChannel(0);
			return true;
		}
		if (id == R.id.action_set_record_channel_first) {
			setRecordChannel(1);
			return true;
		}
		if (id == R.id.action_set_record_channel_second) {
			setRecordChannel(2);
			return true;
		}
		if (id == R.id.action_set_record_channel_summation) {
			setRecordChannel(3);
			return true;
		}
		if (id == R.id.action_set_record_channel_analytic) {
			setRecordChannel(4);
			return true;
		}
		if (id == R.id.action_set_source_default) {
			setAudioSource(MediaRecorder.AudioSource.DEFAULT);
			return true;
		}
		if (id == R.id.action_set_source_microphone) {
			setAudioSource(MediaRecorder.AudioSource.MIC);
			return true;
		}
		if (id == R.id.action_set_source_camcorder) {
			setAudioSource(MediaRecorder.AudioSource.CAMCORDER);
			return true;
		}
		if (id == R.id.action_set_source_voice_recognition) {
			setAudioSource(MediaRecorder.AudioSource.VOICE_RECOGNITION);
			return true;
		}
		if (id == R.id.action_set_source_unprocessed) {
			setAudioSource(MediaRecorder.AudioSource.UNPROCESSED);
			return true;
		}
		if (id == R.id.action_enable_night_mode) {
			AppCompatDelegate.setDefaultNightMode(AppCompatDelegate.MODE_NIGHT_YES);
			return true;
		}
		if (id == R.id.action_disable_night_mode) {
			AppCompatDelegate.setDefaultNightMode(AppCompatDelegate.MODE_NIGHT_NO);
			return true;
		}
		if (id == R.id.action_privacy_policy) {
			showTextPage(getString(R.string.privacy_policy_text));
			return true;
		}
		if (id == R.id.action_about) {
			showTextPage(getString(R.string.about_text, BuildConfig.VERSION_NAME));
			return true;
		}
		return super.onOptionsItemSelected(item);
	}

	private void storeScope() {
		int width = scopeBuffer.width;
		int height = scopeBuffer.height / 2;
		int stride = scopeBuffer.width;
		int offset = stride * scopeBuffer.line;
		storeBitmap(Bitmap.createBitmap(scopeBuffer.pixels, offset, stride, width, height, Bitmap.Config.ARGB_8888));
	}

	private void createScope(Configuration config) {
		int width = scopeBuffer.width;
		int height = scopeBuffer.height / 2;
		DisplayMetrics metrics = getResources().getDisplayMetrics();
		if (config.orientation == Configuration.ORIENTATION_LANDSCAPE)
			height /= 2;
		else
			height = Math.min(Math.max((width * (metrics.heightPixels - 257)) / metrics.widthPixels, height / 2), height);
		scopeBitmap = Bitmap.createBitmap(width, height, Bitmap.Config.ARGB_8888);
		int stride = scopeBuffer.width;
		int offset = stride * (scopeBuffer.line + scopeBuffer.height / 2 - height);
		scopeBitmap.setPixels(scopeBuffer.pixels, offset, stride, 0, 0, width, height);
		scopeView = findViewById(R.id.scope);
		scopeView.setScaleType(ImageView.ScaleType.FIT_CENTER);
		scopeView.setImageBitmap(scopeBitmap);
	}

	private void createFreqPlot(Configuration config) {
		int width = freqPlotBuffer.width;
		int height = freqPlotBuffer.height / 2;
		if (config.orientation != Configuration.ORIENTATION_LANDSCAPE)
			height /= 4;
		freqPlotBitmap = Bitmap.createBitmap(width, height, Bitmap.Config.ARGB_8888);
		int stride = freqPlotBuffer.width;
		int offset = stride * (freqPlotBuffer.line + freqPlotBuffer.height / 2 - height);
		freqPlotBitmap.setPixels(freqPlotBuffer.pixels, offset, stride, 0, 0, width, height);
		freqPlotView = findViewById(R.id.freq_plot);
		freqPlotView.setScaleType(ImageView.ScaleType.FIT_XY);
		freqPlotView.setImageBitmap(freqPlotBitmap);
	}

	private void createPeakMeter() {
		peakMeterBitmap = Bitmap.createBitmap(peakMeterBuffer.width, peakMeterBuffer.height, Bitmap.Config.ARGB_8888);
		peakMeterBitmap.setPixels(peakMeterBuffer.pixels, 0, peakMeterBuffer.width, 0, 0, peakMeterBuffer.width, peakMeterBuffer.height);
		peakMeterView = findViewById(R.id.peak_meter);
		peakMeterView.setScaleType(ImageView.ScaleType.FIT_XY);
		peakMeterView.setImageBitmap(peakMeterBitmap);
	}

	@Override
	public void onConfigurationChanged(@NonNull Configuration config) {
		super.onConfigurationChanged(config);
		setContentView(config.orientation == Configuration.ORIENTATION_LANDSCAPE ? R.layout.activity_main_land : R.layout.activity_main);
		handleInsets();
		createScope(config);
		createFreqPlot(config);
		createPeakMeter();
	}

	private void showTextPage(String message) {
		View view = LayoutInflater.from(this).inflate(R.layout.text_page, null);
		TextView text = view.findViewById(R.id.message);
		text.setText(Html.fromHtml(message, Html.FROM_HTML_MODE_LEGACY));
		text.setMovementMethod(LinkMovementMethod.getInstance());
		AlertDialog.Builder builder = new AlertDialog.Builder(this, R.style.Theme_AlertDialog);
		builder.setNeutralButton(R.string.close, null);
		builder.setView(view);
		builder.show();
	}

	void storeBitmap(Bitmap bitmap) {
		Date date = new Date();
		String name = new SimpleDateFormat("yyyyMMdd_HHmmss", Locale.US).format(date);
		name += ".png";
		String title = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss", Locale.US).format(date);
		ContentValues values = new ContentValues();
		File dir;
		if (Build.VERSION.SDK_INT <= Build.VERSION_CODES.P) {
			dir = Environment.getExternalStoragePublicDirectory(Environment.DIRECTORY_PICTURES);
			if (!dir.exists() && !dir.mkdirs()) {
				showToast(R.string.creating_picture_directory_failed);
				return;
			}
			File file;
			try {
				file = new File(dir, name);
				FileOutputStream stream = new FileOutputStream(file);
				bitmap.compress(Bitmap.CompressFormat.PNG, 100, stream);
				stream.close();
			} catch (IOException e) {
				showToast(R.string.creating_picture_file_failed);
				return;
			}
			values.put(MediaStore.Images.ImageColumns.DATA, file.toString());
		} else {
			values.put(MediaStore.Images.Media.DISPLAY_NAME, name);
			values.put(MediaStore.Images.Media.RELATIVE_PATH, "Pictures/");
			values.put(MediaStore.Images.Media.IS_PENDING, 1);
		}
		values.put(MediaStore.Images.ImageColumns.TITLE, title);
		values.put(MediaStore.Images.ImageColumns.MIME_TYPE, "image/png");
		ContentResolver resolver = getContentResolver();
		Uri uri = resolver.insert(MediaStore.Images.Media.EXTERNAL_CONTENT_URI, values);
		if (uri == null) {
			showToast(R.string.storing_picture_failed);
			return;
		}
		if (Build.VERSION.SDK_INT > Build.VERSION_CODES.P) {
			try {
				ParcelFileDescriptor descriptor = getContentResolver().openFileDescriptor(uri, "w");
				if (descriptor == null) {
					showToast(R.string.storing_picture_failed);
					return;
				}
				FileOutputStream stream = new FileOutputStream(descriptor.getFileDescriptor());
				bitmap.compress(Bitmap.CompressFormat.PNG, 100, stream);
				stream.close();
				descriptor.close();
			} catch (IOException e) {
				showToast(R.string.storing_picture_failed);
				return;
			}
			values.clear();
			values.put(MediaStore.Images.Media.IS_PENDING, 0);
			resolver.update(uri, values, null, null);
		}
		Intent intent = new Intent(Intent.ACTION_SEND);
		intent.putExtra(Intent.EXTRA_STREAM, uri);
		intent.setType("image/png");
		ShareActionProvider share = (ShareActionProvider) MenuItemCompat.getActionProvider(menu.findItem(R.id.menu_item_share));
		if (share != null)
			share.setShareIntent(intent);
		showToast(name);
	}

	private void showToast(String message) {
		Toast toast = Toast.makeText(getApplicationContext(), message, Toast.LENGTH_SHORT);
		toast.setGravity(Gravity.CENTER_HORIZONTAL | Gravity.CENTER_VERTICAL, 0, 0);
		toast.show();
	}

	private void showToast(int id) {
		showToast(getString(id));
	}

	@Override
	protected void onResume() {
		startListening();
		super.onResume();
	}

	@Override
	protected void onPause() {
		stopListening();
		storeSettings();
		super.onPause();
	}
}
